/*
 * Author: Torsten Hilbrich <mail2news@myrkr.in-berlin.de>
 * $Id: mail2news-client.c,v 1.7 2003/10/26 09:54:52 torsten Exp $

 * Small tool to write to the second specified on the command line.
 * If the second argument is given it is written as the first line
 * to the socket.

 * This file is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this software; see the file GPL.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <sys/un.h>
#include <sysexits.h>
#include <signal.h>

const char *socketname = "/var/run/mail2news/mail2news.socket";

void copy(int readFD, int writeFD)
{
  char buffer[1024];
  size_t numBytes;
  size_t bytesWritten;
  size_t offset;

  while(1) {
  again:
    numBytes = read(readFD, buffer, sizeof(buffer));
    /* eof */
    if(numBytes == 0)
      break;

    /* interrupted */
    if(numBytes == -1) {
      if(errno == EINTR)
	goto again;

      fprintf(stderr, "Error while reading input: %s\n",
	      strerror(errno));
      exit(EX_TEMPFAIL);
    }

    offset = 0;
    while(offset < numBytes) {
      bytesWritten = write(writeFD, buffer + offset, numBytes);
      /* bytesWritten == 0 is handled without any special treatment */

      /* interrupted*/
      if(bytesWritten == -1 && errno != EINTR) {
	fprintf(stderr, "Error while writing output: %s\n",
		strerror(errno));
	exit(EX_TEMPFAIL);
      }
      offset += bytesWritten;
      numBytes -= bytesWritten;
    }
  }
}

void timeout(int signo)
{
  fprintf(stderr, "Connection timed out\n");
  exit(EX_TEMPFAIL);
}

int main(int argc, char *argv[]) 
{
  int readFD = 0;
  int sockFD;
  int currentOption;
  int silent = 0;

  struct sockaddr_un sockaddr;

  /* if MAIL2NEWS-SOCKET is set use that value unless a -s is given */
  {
    const char *var = getenv("MAIL2NEWS-SOCKET");
    if(var)
      socketname = var;
  }

  /* check if the first argument is -s or -h */
  currentOption = 1;
  while(currentOption < argc) {
    if(strcmp(argv[currentOption], "-s") == 0) {
      currentOption++;
      if(currentOption < argc) {
	socketname = argv[currentOption];
      } else {
	fprintf(stderr, "Missing socketname argument to -s option\n");
	exit(EX_CONFIG);
      }
    } else if(strcmp(argv[currentOption], "-h") == 0) {
      fprintf(stdout, "Usage: %s [ -q ] [ -s socketname ] [ newsgroup ... ]\n",
	      argv[0]);
      exit(0);
    } else if(strcmp(argv[currentOption], "-q") == 0) {
      silent = 1;
      currentOption++;
    } else {
      break;
    }
  }

  sockFD = socket(PF_LOCAL, SOCK_STREAM, 0);
  if(sockFD == -1) {
    perror("socket");
    exit(EX_TEMPFAIL);
  }

  memset(&sockaddr, 0, sizeof(sockaddr));
  sockaddr.sun_family = AF_LOCAL;
  if(strlen(socketname) >= 104) {
    fprintf(stderr, "%s: %d > %d\n", "Socket file name too long",
	    strlen(socketname), 104);
    exit(EX_CONFIG);
  }

  strcpy(sockaddr.sun_path, socketname);

  /* set a timeout of 10 seconds*/
  signal(SIGALRM, timeout);
  alarm(10);  

  if(connect(sockFD, (struct sockaddr *)&sockaddr, SUN_LEN(&sockaddr)) < 0) {
    perror("connect");
    exit(EX_TEMPFAIL);
  }

  while(currentOption < argc) {
    write(sockFD, argv[currentOption], strlen(argv[currentOption]));
    write(sockFD, "\n", 1);
    currentOption++;
  }
  write(sockFD, "\n", 1);

  copy(readFD, sockFD);
  shutdown(sockFD, 1);

  if(! silent)
  {
    char buffer[1024];
    int offset = 0;
    char *endptr;
    long result;
    int numBytes = sizeof(buffer) - 1;
    
    /* just in case the while statement gets true */
    buffer[numBytes] = 0;

    while(offset < numBytes) {
      int bytesRead = read(sockFD, buffer + offset, numBytes);
      if(bytesRead == -1 && errno == EINTR)
	continue;
      if(bytesRead == 0) {
	/* append \0 */
	buffer[offset] = 0;
	break;
      }
      offset += bytesRead;
      numBytes -= bytesRead;
    }

    if(buffer[0] == 0) {
      fprintf(stderr, "Unable to read the server's reply\n");
      exit (EX_TEMPFAIL);
    }

    result = strtol(buffer, &endptr, 10);
    if(*endptr == 0) {
      /* everything okay */

      exit((int)result);
    } else {
      fprintf(stderr, "Server reply %s is no valid number\n",
	      buffer);
      exit (EX_TEMPFAIL);
    }
  }
  alarm(0);
  return EXIT_SUCCESS;
}
